# Copyright 2022 J.P. Morgan Chase & Co.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import datetime
import uuid

import py_avro_schema as pas
from py_avro_schema._testing import assert_schema


def test_date():
    py_type = datetime.date
    expected = {
        "type": "int",
        "logicalType": "date",
    }
    assert_schema(py_type, expected)


def test_time():
    py_type = datetime.time
    expected = {
        "type": "long",
        "logicalType": "time-micros",
    }
    assert_schema(py_type, expected)


def test_time_milliseconds():
    py_type = datetime.time
    expected = {
        "type": "int",
        "logicalType": "time-millis",
    }
    options = pas.Option.MILLISECONDS
    assert_schema(py_type, expected, options=options)


def test_datetime():
    py_type = datetime.datetime
    expected = {
        "type": "long",
        "logicalType": "timestamp-micros",
    }
    assert_schema(py_type, expected)


def test_datetime_milliseconds():
    py_type = datetime.datetime
    expected = {
        "type": "long",
        "logicalType": "timestamp-millis",
    }
    options = pas.Option.MILLISECONDS
    assert_schema(py_type, expected, options=options)


def test_timedelta():
    py_type = datetime.timedelta
    expected = {
        "type": "fixed",
        "name": "datetime.timedelta",
        "size": 12,
        "logicalType": "duration",
    }
    assert_schema(py_type, expected)


def test_decimal():
    py_type = pas.DecimalType[5, 2]
    expected = {
        "type": "bytes",
        "logicalType": "decimal",
        "precision": 5,
        "scale": 2,
    }
    assert_schema(py_type, expected)


def test_multiple_decimals():
    # Test the magic with _GenericAlias!
    py_type_1 = pas.DecimalType[5, 2]
    expected_1 = {
        "type": "bytes",
        "logicalType": "decimal",
        "precision": 5,
        "scale": 2,
    }
    py_type_2 = pas.DecimalType[3, 1]
    expected_2 = {
        "type": "bytes",
        "logicalType": "decimal",
        "precision": 3,
        "scale": 1,
    }
    assert_schema(py_type_1, expected_1)
    assert_schema(py_type_2, expected_2)


def test_uuid():
    py_type = uuid.UUID
    expected = {
        "type": "string",
        "logicalType": "uuid",
    }
    assert_schema(py_type, expected)



# Copyright 2022 J.P. Morgan Chase & Co.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


"""
Generate Apache Avro schemas for Python types including standard library data-classes and Pydantic data models.

The main API is a single function, :func:`generate`. Its first argument is the Python type or class to generate the Avro
schema for.

.. seealso::

   Data types supported by **py-avro-schema**: :doc:`types`.

"""

from typing import Optional, Type

import memoization
import orjson

from py_avro_schema._schemas import JSON_OPTIONS, Option, schema
from py_avro_schema._typing import DecimalType

try:
    from importlib import metadata
except ImportError:  # pragma: no cover
    # Python < 3.8
    import importlib_metadata as metadata  # type: ignore

#: Library version, e.g. 1.0.0, taken from Git tags
__version__ = metadata.version("py-avro-schema")


__all__ = [
    "generate",
    "DecimalType",
    "Option",
]


@memoization.cached
def generate(
    py_type: Type,
    *,
    namespace: Optional[str] = None,
    options: Option = Option(0),
) -> bytes:
    """
    Return an Avro schema as a JSON-formatted bytestring for a given Python class or instance

    This function is cached and can be called repeatedly with the same arguments without any performance penalty.

    :param py_type:   The Python class to generate a schema for.
    :param namespace: The Avro namespace to add to schemas.
    :param options:   Schema generation options as defined by :class:`Option` enum values. Specify multiple values like
                      this: ``Option.INT_32 | Option.FLOAT_32``.
    """
    schema_dict = schema(py_type, namespace=namespace, options=options)
    json_options = 0
    for opt in JSON_OPTIONS:
        if opt in options:
            json_options |= opt.value
    schema_json = orjson.dumps(schema_dict, option=json_options)
    return schema_json



# Copyright 2022 J.P. Morgan Chase & Co.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re

import pytest

from py_avro_schema._testing import assert_schema


def test_plain_class_with_type_hints():
    class PyType:
        """A port, not using dataclass"""

        def __init__(
            self, name: str, *, country: str = "NLD", latitude: float, longitude: float
        ):  # note the non-default kwargs near the end!
            self.name = name
            self.country = country.upper()
            self.latitude = latitude
            self.longitude = longitude

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "name",
                "type": "string",
            },
            {
                "name": "country",
                "type": "string",
                "default": "NLD",
            },
            {
                "name": "latitude",
                "type": "double",
            },
            {
                "name": "longitude",
                "type": "double",
            },
        ],
    }

    assert_schema(PyType, expected)


def test_plain_class_no_type_hints():
    class PyType:
        """A port, not using dataclass"""

        def __init__(self, name, *, country="NLD", latitude, longitude):  # note the non-default kwargs near the end!
            self.name = name
            self.country = country.upper()
            self.latitude = latitude
            self.longitude = longitude

    with pytest.raises(
        TypeError,
        match=re.escape(
            "Cannot generate Avro schema for Python type <class 'test_plain_class.test_plain_class_no_type_hints."
            "<locals>.PyType'>"
        ),
    ):
        assert_schema(PyType, {})



# Copyright 2022 J.P. Morgan Chase & Co.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


"""
Test functions
"""
import dataclasses
import difflib
from typing import Dict, Type, Union

import avro.schema  # type: ignore
import orjson

import py_avro_schema._schemas


def assert_schema(py_type: Type, expected_schema: Union[str, Dict[str, str]], **kwargs) -> None:
    """Test that the given Python type results in the correct Avro schema"""
    if not kwargs.pop("do_auto_namespace", False):
        kwargs["options"] = kwargs.get("options", py_avro_schema.Option(0)) | py_avro_schema.Option.NO_AUTO_NAMESPACE
    if not kwargs.pop("do_doc", False):
        kwargs["options"] = kwargs.get("options", py_avro_schema.Option(0)) | py_avro_schema.Option.NO_DOC
    actual_schema = py_avro_schema._schemas.schema(py_type, **kwargs)
    expected_schema_json = orjson.dumps(expected_schema, option=orjson.OPT_INDENT_2 | orjson.OPT_SORT_KEYS).decode()
    actual_schema_json = orjson.dumps(actual_schema, option=orjson.OPT_INDENT_2 | orjson.OPT_SORT_KEYS).decode()
    if actual_schema != expected_schema:
        print("Expected schema:")
        print(expected_schema_json)
        print("Actual schema:")
        print(actual_schema_json)
        print("Differences:")
        for diff in difflib.unified_diff(
            expected_schema_json.splitlines(),
            actual_schema_json.splitlines(),
            fromfile="expected",
            tofile="actual",
            n=5,
        ):
            print(diff)

    assert actual_schema == expected_schema
    # Assert that we can parse the schema data as a valid Avro schema
    assert avro.schema.make_avsc_object(actual_schema, None)


@dataclasses.dataclass
class PyType:
    """For testing"""

    field_a: str



# Copyright 2022 J.P. Morgan Chase & Co.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import uuid
from typing import List, Optional

import pydantic
import pytest

from py_avro_schema._testing import assert_schema


def test_string_field():
    class PyType(pydantic.BaseModel):
        field_a: str

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": "string",
            }
        ],
    }
    assert_schema(PyType, expected)


def test_string_field_default():
    class PyType(pydantic.BaseModel):
        field_a: str = ""

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": "string",
                "default": "",
            }
        ],
    }
    assert_schema(PyType, expected)


def test_string_field_default_wrong_type():
    class PyType(pydantic.BaseModel):
        field_a: str = 1  # That's not valid, because field type is str

    with pytest.raises(TypeError, match="type of field field_a must be str; got int instead"):
        assert_schema(PyType, {})


def test_optional_field_default():
    class PyType(pydantic.BaseModel):
        field_a: Optional[str] = None

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": [
                    "null",
                    "string",
                ],
                "default": None,
            }
        ],
    }
    assert_schema(PyType, expected)


def test_optional_field_default_2():
    class PyType(pydantic.BaseModel):
        field_a: str = None  # In Pydantic you can do this :(

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": [
                    "null",
                    "string",
                ],
                "default": None,
            }
        ],
    }
    assert_schema(PyType, expected)


def test_list_string_field():
    class PyType(pydantic.BaseModel):
        field_a: List[str]

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": {
                    "type": "array",
                    "items": "string",
                },
            },
        ],
    }
    assert_schema(PyType, expected)


def test_list_string_field_default():
    class PyType(pydantic.BaseModel):
        field_a: List[str] = []  # Pydantic allows mutable defaults like this

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": {
                    "type": "array",
                    "items": "string",
                },
                "default": [],
            },
        ],
    }
    assert_schema(PyType, expected)


def test_list_string_field_default_wrong_type():
    class PyType(pydantic.BaseModel):
        field_a: List[str] = [1]  # Pydantic allows mutable defaults like this

    with pytest.raises(TypeError, match=re.escape("type of field field_a[0] must be str; got int instead")):
        assert_schema(PyType, {})


def test_field():
    class PyTypeChild(pydantic.BaseModel):
        field_a: str

    class PyType(pydantic.BaseModel):
        field_child: PyTypeChild

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_child",
                "type": {
                    "type": "record",
                    "name": "PyTypeChild",
                    "fields": [
                        {
                            "name": "field_a",
                            "type": "string",
                        },
                    ],
                },
            },
        ],
    }
    assert_schema(PyType, expected)


def test_list_dataclass_field():
    class PyTypeChild(pydantic.BaseModel):
        field_a: str

    class PyType(pydantic.BaseModel):
        field_child: List[PyTypeChild]

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_child",
                "type": {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "PyTypeChild",
                        "fields": [
                            {
                                "name": "field_a",
                                "type": "string",
                            },
                        ],
                    },
                },
            },
        ],
    }
    assert_schema(PyType, expected)


def test_repeated_field():
    class PyTypeChild(pydantic.BaseModel):
        field_a: str

    class PyType(pydantic.BaseModel):
        field_child_1: PyTypeChild
        field_child_2: PyTypeChild

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_child_1",
                "type": {
                    "type": "record",
                    "name": "PyTypeChild",
                    "fields": [
                        {
                            "name": "field_a",
                            "type": "string",
                        }
                    ],
                },
            },
            {
                "name": "field_child_2",
                "type": "PyTypeChild",
            },
        ],
    }
    assert_schema(PyType, expected)


def test_self_ref_field():
    class PyType(pydantic.BaseModel):
        field_a: "PyType"

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": "PyType",
            },
        ],
    }
    assert_schema(PyType, expected)


def test_class_docs():
    class PyType(pydantic.BaseModel):
        """My PyType"""

        field_a: str

    expected = {
        "type": "record",
        "name": "PyType",
        "doc": "My PyType",
        "fields": [
            {
                "name": "field_a",
                "type": "string",
            }
        ],
    }
    assert_schema(PyType, expected, do_doc=True)


def test_field_docs():
    class PyType(pydantic.BaseModel):
        """My PyType"""

        field_a: str = pydantic.Field(description="My field_a")

    expected = {
        "type": "record",
        "name": "PyType",
        "doc": "My PyType",
        "fields": [
            {
                "name": "field_a",
                "type": "string",
                "doc": "My field_a",
            }
        ],
    }
    assert_schema(PyType, expected, do_doc=True)


def test_uuid():
    py_type = pydantic.UUID4
    expected = {
        "type": "string",
        "logicalType": "uuid",
    }
    assert_schema(py_type, expected)


def test_uuid_field_default():
    class PyType(pydantic.BaseModel):
        # Using :mod:`uuid` as the library to generate the ID, then cast to :class:`pydantic.UUID4` to match the type
        # hint. Alternatively, just use plain :class:`uuid.UUID` as the type hint.
        field_a: pydantic.UUID4 = pydantic.Field(default_factory=lambda: pydantic.UUID4(uuid.uuid4().hex))

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": {
                    "type": "string",
                    "logicalType": "uuid",
                },
                "default": "",
            },
        ],
    }
    assert_schema(PyType, expected)


def test_positive_float_field():
    class PyType(pydantic.BaseModel):
        field_a: pydantic.PositiveFloat

    expected = {
        "type": "record",
        "name": "PyType",
        "fields": [
            {
                "name": "field_a",
                "type": "double",
            }
        ],
    }
    assert_schema(PyType, expected)
